"""Convenience Module for creating different c3_backend
"""
from typing import Dict, List
import tensorflow as tf
import math
from .c3_exceptions import C3QiskitError

GATE_MAP = {
    "X": "RXp",
    "x": "RXp",
    "Y": "RYp",
    "y": "RYp",
    "Z": "RZp",
    "z": "RZp",
    "CNOT": "CRXp",
    "CX": "CRXp",
    "cx": "CRXp",
    "CZ": "CRZp",
    "cz": "CRZp",
    "I": "Id",
    "u0": "Id",
    "id": "Id",
    "RX90p": "RX90p",  # RX((np.pi)/2)
    "RX90m": "RX90m",  # RX((-np.pi)/2)
    "RY90p": "RY90p",  # RY((np.pi)/2)
    "RY90m": "RY90m",  # RY(-(np.pi)/2)
    "RZ90p": "RZ90p",  # RZ((np.pi)/2)
    "RZ90m": "RZ90m",  # RZ(-(np.pi)/2)
    "iSwap": "ISWAP",
}


def pad_gate_name(gate_name: str, qubits: List[int], n_qubits: int) -> str:
    """Pad gate name with Identity gates in correct indices

    Parameters
    ----------
    gate_name : str
        A C3 compatible gate name
    qubits : List[int]
        Indices to apply gate
    n_qubits : int
        Total number of qubits in the device

    Returns
    -------
    str
        Identity padded gate name, eg ::

            pad_gate_name("CCRX90p", [1, 2, 3], 5) -> 'Id:CCRX90p:Id'
            pad_gate_name("RX90p", [0], 5) -> 'RX90p:Id:Id:Id:Id'
    """

    # TODO (check) Assumption control and action qubits next to each other
    gate_names = ["Id"] * (n_qubits - (len(qubits) - 1))
    gate_names[qubits[0]] = gate_name
    padded_gate_str = ":".join(gate_names)
    return padded_gate_str


def get_sequence(instructions: List, n_qubits: int) -> List[str]:
    """Return a sequence of gates from instructions

    Parameters
    ----------
    instructions : List[dict]
        Instructions from the qasm experiment, for example::

        instructions: [
                {"name": "u1", "qubits": [0], "params": [0.4]},
                {"name": "u2", "qubits": [0], "params": [0.4,0.2]},
                {"name": "u3", "qubits": [0], "params": [0.4,0.2,-0.3]},
                {"name": "snapshot", "label": "snapstate1", "snapshot_type": "statevector"},
                {"name": "cx", "qubits": [0,1]},
                {"name": "barrier", "qubits": [0]},
                {"name": "measure", "qubits": [0], "register": [1], "memory": [0]},
                {"name": "u2", "qubits": [0], "params": [0.4,0.2], "conditional": 2}
            ]

    n_qubits: int
        Number of qubits in the device config

    Returns
    -------
    List[str]
        List of gates, for example::

        sequence = ["RX90p:Id", "Id:RX90p", "CR90"]

    """

    sequence = []

    for instruction in instructions:

        # TODO Check if gate is possible from device_config
        # TODO parametric gates

        iname = instruction.name
        # Conditional operations are not supported
        conditional = getattr(instructions, "conditional", None)  # noqa
        if conditional is not None:
            raise C3QiskitError("C3 Simulator does not support conditional operations")

        # reset, binary functions is not supported
        elif iname in ["reset", "bfunc"]:
            raise C3QiskitError("C3 Simulator does not support {}".format(iname))

        # barrier is implemented internally through Identity gates
        elif iname == "barrier":
            pass

        # TODO U, u3
        elif iname in ("U", "u3"):
            raise C3QiskitError("U3 gates are not yet implemented")

        # measure implemented outside sequences
        elif iname == "measure":
            pass

        elif iname in GATE_MAP.keys():
            gate_name = GATE_MAP[iname]
            qubits = instruction.qubits
            padded_gate_str = pad_gate_name(gate_name, qubits, n_qubits)
            sequence.append(padded_gate_str)

        # raise C3QiskitError if unknown instruction
        else:
            raise C3QiskitError("Encountered unknown operation {}".format(iname))

    return sequence


def get_init_ground_state(n_qubits: int, n_levels: int) -> tf.Tensor:
    """Return a perfect ground state

    Parameters
    ----------
    n_qubits : int
        Number of qubits in the system

    n_levels : int
        Number of levels for each qubit

    Returns
    -------
    tf.Tensor
        Tensor array of ground state
        shape(m^n, 1), dtype=complex128
        m = no of qubit levels
        n = no of qubits
    """
    psi_init = [[0] * (int)(math.pow(n_levels, n_qubits))]
    psi_init[0][0] = 1
    init_state = tf.transpose(tf.constant(psi_init, tf.complex128))

    return init_state


def flip_labels(counts: Dict[str, int]) -> Dict[str, int]:
    """Flip C3 qubit labels to match Qiskit qubit indexing

    Parameters
    ----------
    counts : Dict[str, int]
        OpenQasm 2.0 result counts with original C3 style
        qubit indices

    Returns
    -------
    Dict[str, int]
        OpenQasm 2.0 result counts with Qiskit style labels

    Note
    ----
    Basis vector ordering in Qiskit

    Qiskit uses a slightly different ordering of the qubits compared to
    what is seen in Physics textbooks. In qiskit, the qubits are represented from
    the most significant bit (MSB) on the left to the least significant bit (LSB)
    on the right (big-endian). This is similar to bitstring representation
    on classical computers, and enables easy conversion from bitstrings to
    integers after measurements are performed.

    More details:
    https://qiskit.org/documentation/tutorials/circuits/3_summary_of_quantum_operations.html#Basis-vector-ordering-in-Qiskit

    """
    labels_flipped_counts = {}
    for key, value in counts.items():
        key_bin = bin(int(key, 0))
        key_bin_rev = "0b" + key_bin[:1:-1]
        key_rev = hex(int(key_bin_rev, 0))
        labels_flipped_counts[key_rev] = value
    return labels_flipped_counts



"""Module for testing C3 Integration with Qiskit
"""

from c3.qiskit import C3Provider
from c3.qiskit.c3_exceptions import C3QiskitError
from qiskit.quantum_info import Statevector
from qiskit import transpile
from qiskit.providers import BackendV1 as Backend
from qiskit import execute, Aer
from qiskit.transpiler.exceptions import TranspilerError

import pytest


@pytest.mark.unit
@pytest.mark.qiskit
def test_backends():
    """Test backends() function which returns all available backends"""
    c3_qiskit = C3Provider()
    for backend in c3_qiskit.backends():
        assert isinstance(backend, Backend)


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.parametrize("backend", ["c3_qasm_perfect_simulator"])
def test_get_backend(backend):
    """Test get_backend() which returns the backend with matching name

    Parameters
    ----------
    backend : str
        name of the backend that is to be fetched
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    assert isinstance(received_backend, Backend)


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.xfail(raises=TranspilerError)
@pytest.mark.parametrize("backend", ["c3_qasm_perfect_simulator"])
def test_transpile(get_test_circuit, backend):  # noqa
    """Test the transpiling using our backends.
    Should throw error due to missing H gate

    Parameters
    ----------
    get_test_circuit : callable
        pytest fixture for a simple quantum circuit
    backend : str
        name of the backend which is to be tested
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    trans_qc = transpile(get_test_circuit, received_backend)
    assert Statevector.from_instruction(get_test_circuit).equiv(
        Statevector.from_instruction(trans_qc)
    )


@pytest.mark.integration
@pytest.mark.qiskit
@pytest.mark.slow
@pytest.mark.parametrize("backend", ["c3_qasm_perfect_simulator"])
def test_get_result(get_6_qubit_circuit, backend, get_result_qiskit):  # noqa
    """Test the counts from running a 6 qubit Circuit

    Parameters
    ----------
    get_6_qubit_circuit : callable
        pytest fixture for a 6 qubit circuit
    backend : str
        name of the backend which is to be tested
    simulation_type: str
        physics based or perfect gates simulation
    get_result_qiskit: callable
        pytest fixture for experiment results
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    received_backend.set_device_config("test/quickstart.hjson")
    qc = get_6_qubit_circuit
    job_sim = execute(qc, received_backend, shots=1000)
    result_sim = job_sim.result()

    # Test results with qiskit style qubit indexing
    qiskit_simulator = Aer.get_backend("qasm_simulator")
    qiskit_counts = execute(qc, qiskit_simulator, shots=1000).result().get_counts(qc)
    assert result_sim.get_counts(qc) == qiskit_counts

    # Test results with original C3 style qubit indexing
    received_backend.disable_flip_labels()
    no_flip_counts = get_result_qiskit[backend]
    job_sim = execute(qc, received_backend, shots=1000)
    result_sim = job_sim.result()
    assert result_sim.get_counts() == no_flip_counts


@pytest.mark.unit
@pytest.mark.qiskit
@pytest.mark.xfail(raises=C3QiskitError)
@pytest.mark.slow
@pytest.mark.parametrize("backend", ["c3_qasm_perfect_simulator"])
def test_get_exception(get_bad_circuit, backend):  # noqa
    """Test to check exceptions

    Parameters
    ----------
    get_bad_circuit : callable
        pytest fixture for a circuit with conditional operation
    backend : str
        name of the backend which is to be tested
    """
    c3_qiskit = C3Provider()
    received_backend = c3_qiskit.get_backend(backend)
    received_backend.set_device_config("test/quickstart.hjson")
    qc = get_bad_circuit
    job_sim = execute(qc, received_backend, shots=1000)
    result_sim = job_sim.result()
    assert result_sim.get_counts(qc) == {}



"""Library of fidelity functions."""
# TODO think of how to use the fidelity functions in a cleaner way

import numpy as np
import tensorflow as tf

# import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from c3.utils.tf_utils import (
    tf_ave,
    tf_super,
    tf_abs,
    tf_ketket_fid,
    tf_superoper_unitary_overlap,
    tf_unitary_overlap,
    tf_dm_to_vec,
    tf_average_fidelity,
    tf_superoper_average_fidelity,
    tf_state_to_dm,
    evaluate_sequences,
)
from c3.utils.qt_utils import (
    basis,
    perfect_gate,
    perfect_cliffords,
    cliffords_decomp,
    cliffords_decomp_xId,
    single_length_RB,
    cliffords_string,
    projector,
)

fidelities = dict()


def fid_reg_deco(func):
    """
    Decorator for making registry of functions
    """
    fidelities[str(func.__name__)] = func
    return func


@fid_reg_deco
def state_transfer_infid_set(U_dict: dict, index, dims, psi_0, proj=True):
    """
    Mean state transfer infidelity.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    psi_0 : tf.Tensor
        Initial state of the device
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float
        State infidelity, averaged over the gates in U_dict
    """
    infids = []
    for gate in U_dict.keys():
        infid = state_transfer_infid(U_dict, gate, index, dims, psi_0, proj)
        infids.append(infid)
    return tf.reduce_mean(infids)


def state_transfer_infid(U_dict: dict, gate: str, index, dims, psi_0, proj: bool):
    """
    Single gate state transfer infidelity.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    gate : str
        One of the keys of U_dict, selects the gate to be evaluated
    dims : list
        List of dimensions of qubits
    psi_0 : tf.Tensor
        Initial state of the device
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float
        State infidelity for the selected gate

    """
    U = U_dict[gate]
    projection = "fulluni"
    if proj:
        projection = "wzeros"
    U_ideal = tf.constant(
        perfect_gate(gate, index, dims, projection), dtype=tf.complex128
    )
    psi_ideal = tf.matmul(U_ideal, psi_0)
    psi_actual = tf.matmul(U, psi_0)
    overlap = tf_ketket_fid(psi_ideal, psi_actual)
    infid = 1 - overlap
    return infid


@fid_reg_deco
def unitary_infid(U_dict: dict, gate: str, index, dims, proj: bool):
    """
    Unitary overlap between ideal and actually performed gate.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    gate : str
        One of the keys of U_dict, selects the gate to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float
        Unitary fidelity.
    """
    U = U_dict[gate]
    projection = "fulluni"
    fid_lvls = np.prod([dims[i] for i in index])
    if proj:
        projection = "wzeros"
        fid_lvls = 2 ** len(index)
    U_ideal = tf.constant(
        perfect_gate(gate, index, dims, projection), dtype=tf.complex128
    )
    infid = 1 - tf_unitary_overlap(U, U_ideal, lvls=fid_lvls)
    return infid


@fid_reg_deco
def unitary_infid_set(U_dict: dict, index, dims, eval, proj=True):
    """
    Mean unitary overlap between ideal and actually performed gate for the gates in
    U_dict.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float
        Unitary fidelity.
    """
    infids = []
    for gate in U_dict.keys():
        infid = unitary_infid(U_dict, gate, index, dims, proj)
        infids.append(infid)
    return tf.reduce_mean(infids)


@fid_reg_deco
def lindbladian_unitary_infid(U_dict: dict, gate: str, index, dims, proj: bool):
    """
    Variant of the unitary fidelity for the Lindbladian propagator.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    gate : str
        One of the keys of U_dict, selects the gate to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float
        Overlap fidelity for the Lindblad propagator.
    """
    # Here we deal with the projected case differently because it's not easy
    # to select the right section of the superoper
    U = U_dict[gate]
    projection = "fulluni"
    fid_lvls = np.prod([dims[i] for i in index])
    if proj:
        projection = "wzeros"
        fid_lvls = 2 ** len(index)
    U_ideal = tf_super(
        tf.constant(perfect_gate(gate, index, dims, projection), dtype=tf.complex128)
    )
    infid = 1 - tf_superoper_unitary_overlap(U, U_ideal, lvls=fid_lvls)
    return infid


@fid_reg_deco
def lindbladian_unitary_infid_set(U_dict: dict, index, dims, eval, proj=True):
    """
    Variant of the mean unitary fidelity for the Lindbladian propagator.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float
        Mean overlap fidelity for the Lindblad propagator for all gates in U_dict.
    """
    infids = []
    for gate in U_dict.keys():
        infid = lindbladian_unitary_infid(U_dict, gate, index, dims, proj)
        infids.append(infid)
    return tf.reduce_mean(infids)


@fid_reg_deco
def average_infid_CZ(U_dict: dict, index, dims, eval, proj=True):
    """
    Average fidelity uses the Pauli basis to compare. Thus, perfect gates are
    always 2x2 (per qubit) and the actual unitary needs to be projected down.
    Variant for two-qubit gates.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace
    """
    proj = projector(dims, index)
    U = proj @ U_dict["Id:CRZp"] @ proj.T
    subspace_dims = [dims[index[0]], dims[index[1]]]
    U_ideal = tf.constant(perfect_gate("CRZp", index=[0, 1], dims=[2, 2]))
    infid = 1 - tf_average_fidelity(U, U_ideal, lvls=subspace_dims)
    return infid


@fid_reg_deco
def average_infid_simult(U_dict: dict, gate: str, index, dims, proj=True):
    """
    Average fidelity uses the Pauli basis to compare. Thus, perfect gates are
    always 2x2 (per qubit) and the actual unitary needs to be projected down.
    Variant for simultaneous single qubit gates.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace
    """
    proj = projector(dims, index)
    U = proj @ U_dict[gate] @ proj.T
    gate_split = gate.split(":")
    two_qubit_gate = ":".join([gate_split[index[0]], gate_split[index[1]]])
    subspace_dims = [dims[index[0]], dims[index[1]]]
    U_ideal = tf.constant(perfect_gate(two_qubit_gate, index=[0, 1], dims=[2, 2]))
    infid = 1 - tf_average_fidelity(U, U_ideal, lvls=subspace_dims)
    return infid


@fid_reg_deco
def average_infid(U_dict: dict, gate: str, index, dims, proj=True):
    """
    Average fidelity uses the Pauli basis to compare. Thus, perfect gates are
    always 2x2 (per qubit) and the actual unitary needs to be projected down.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace
    """
    U = U_dict[gate]
    U_ideal = tf.constant(
        perfect_gate(gate, index, dims=[2] * len(dims)), dtype=tf.complex128
    )
    infid = 1 - tf_average_fidelity(U, U_ideal, lvls=dims)
    return infid


@fid_reg_deco
def average_infid_set(U_dict: dict, index, dims, eval, proj=True):
    """
    Mean average fidelity over all gates in U_dict.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float64
        Mean average fidelity
    """
    infids = []
    for gate in U_dict.keys():
        infid = average_infid(U_dict, gate, index, dims, proj)
        infids.append(infid)
    return tf.reduce_mean(infids)


@fid_reg_deco
def average_infid_seq(U_dict: dict, index, dims, eval, proj=True):
    """
    Average sequence fidelity over all gates in U_dict.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float64
        Mean average fidelity
    """
    fid = 1
    for gate in U_dict.keys():
        fid *= 1 - average_infid(U_dict, gate, index, dims, proj)
    return 1 - fid


@fid_reg_deco
def lindbladian_average_infid(U_dict: dict, gate: str, index, dims, proj=True):
    """
    Average fidelity uses the Pauli basis to compare. Thus, perfect gates are
    always 2x2 (per qubit) and the actual unitary needs to be projected down.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace
    """
    U = U_dict[gate]
    ideal = tf.constant(
        perfect_gate(gate, index, dims=[2] * len(dims)), dtype=tf.complex128
    )
    U_ideal = tf_super(ideal)
    infid = 1 - tf_superoper_average_fidelity(U, U_ideal, lvls=dims)
    return infid


@fid_reg_deco
def lindbladian_average_infid_set(U_dict: dict, index, dims, eval, proj=True):
    """
    Mean average fidelity over all gates in U_dict.

    Parameters
    ----------
    U_dict : dict
        Contains unitary representations of the gates, identified by a key.
    index : int
        Index of the qubit(s) in the Hilbert space to be evaluated
    dims : list
        List of dimensions of qubits
    proj : boolean
        Project to computational subspace

    Returns
    -------
    tf.float64
        Mean average fidelity
    """
    infids = []
    for gate in U_dict.keys():
        infid = lindbladian_average_infid(U_dict, gate, index, dims, proj)
        infids.append(infid)
    return tf.reduce_mean(infids)


@fid_reg_deco
def epc_analytical(U_dict: dict, index, dims, proj: bool, cliffords=False):
    # TODO check this work with new index and dims (double-check)
    num_gates = len(dims)
    if cliffords:
        real_cliffords = evaluate_sequences(U_dict, [[C] for C in cliffords_string])
    elif num_gates == 1:
        real_cliffords = evaluate_sequences(U_dict, cliffords_decomp)
    elif num_gates == 2:
        real_cliffords = evaluate_sequences(U_dict, cliffords_decomp_xId)
    ideal_cliffords = perfect_cliffords(lvls=[2] * num_gates, num_gates=num_gates)
    fids = []
    for C_indx in range(24):
        C_real = real_cliffords[C_indx]
        C_ideal = tf.constant(ideal_cliffords[C_indx], dtype=tf.complex128)
        ave_fid = tf_average_fidelity(C_real, C_ideal, lvls=dims)
        fids.append(ave_fid)
    infid = 1 - tf_ave(fids)
    return infid


@fid_reg_deco
def lindbladian_epc_analytical(U_dict: dict, index, dims, proj: bool, cliffords=False):
    num_gates = len(dims)
    if cliffords:
        real_cliffords = evaluate_sequences(U_dict, [[C] for C in cliffords_string])
    elif num_gates == 1:
        real_cliffords = evaluate_sequences(U_dict, cliffords_decomp)
    elif num_gates == 2:
        real_cliffords = evaluate_sequences(U_dict, cliffords_decomp_xId)
    ideal_cliffords = perfect_cliffords(lvls=[2] * num_gates, num_gates=num_gates)
    fids = []
    for C_indx in range(24):
        C_real = real_cliffords[C_indx]
        C_ideal = tf_super(tf.constant(ideal_cliffords[C_indx], dtype=tf.complex128))
        ave_fid = tf_superoper_average_fidelity(C_real, C_ideal, lvls=dims)
        fids.append(ave_fid)
    infid = 1 - tf_ave(fids)
    return infid


@fid_reg_deco
def populations(state, lindbladian):
    if lindbladian:
        diag = []
        dim = int(np.sqrt(len(state)))
        indeces = [n * dim + n for n in range(dim)]
        for indx in indeces:
            diag.append(state[indx])
        return np.abs(diag)
    else:
        return np.abs(state) ** 2


@fid_reg_deco
def population(U_dict: dict, lvl: int, gate: str):
    U = U_dict[gate]
    lvls = U.shape[0]
    psi_0 = tf.constant(basis(lvls, 0), dtype=tf.complex128)
    psi_actual = tf.matmul(U, psi_0)
    return populations(psi_actual, lindbladian=False)[lvl]


def lindbladian_population(U_dict: dict, lvl: int, gate: str):
    U = U_dict[gate]
    lvls = int(np.sqrt(U.shape[0]))
    psi_0 = tf.constant(basis(lvls, 0), dtype=tf.complex128)
    dv_0 = tf_dm_to_vec(tf_state_to_dm(psi_0))
    dv_actual = tf.matmul(U, dv_0)
    return populations(dv_actual, lindbladian=True)[lvl]


@fid_reg_deco
def RB(
    U_dict,
    min_length: int = 5,
    max_length: int = 500,
    num_lengths: int = 20,
    num_seqs: int = 30,
    logspace=False,
    lindbladian=False,
    padding="",
):
    gate = list(U_dict.keys())[0]
    U = U_dict[gate]
    dim = int(U.shape[0])
    psi_init = tf.constant(basis(dim, 0), dtype=tf.complex128)
    if logspace:
        lengths = np.rint(
            np.logspace(np.log10(min_length), np.log10(max_length), num=num_lengths)
        ).astype(int)
    else:
        lengths = np.rint(np.linspace(min_length, max_length, num=num_lengths)).astype(
            int
        )
    surv_prob = []
    for L in lengths:
        seqs = single_length_RB(num_seqs, L, padding)
        Us = evaluate_sequences(U_dict, seqs)
        pop0s = []
        for U in Us:
            pops = populations(tf.matmul(U, psi_init), lindbladian)
            pop0s.append(float(pops[0]))
        surv_prob.append(pop0s)

    def RB_fit(len, r, A, B):
        return A * r ** (len) + B

    bounds = (0, 1)
    init_guess = [0.9, 0.5, 0.5]
    fitted = False
    while not fitted:
        try:
            means = np.mean(surv_prob, axis=1)
            stds = np.std(surv_prob, axis=1) / np.sqrt(len(surv_prob[0]))
            solution, cov = curve_fit(
                RB_fit, lengths, means, sigma=stds, bounds=bounds, p0=init_guess
            )
            r, A, B = solution
            fitted = True
        except Exception as message:
            print(message)
            if logspace:
                new_lengths = np.rint(
                    np.logspace(
                        np.log10(max_length + min_length),
                        np.log10(max_length * 2),
                        num=num_lengths,
                    )
                ).astype(int)
            else:
                new_lengths = np.rint(
                    np.linspace(
                        max_length + min_length, max_length * 2, num=num_lengths
                    )
                ).astype(int)
            max_length = max_length * 2
            for L in new_lengths:
                seqs = single_length_RB(num_seqs, L, padding)
                Us = evaluate_sequences(U_dict, seqs)
                pop0s = []
                for U in Us:
                    pops = populations(tf.matmul(U, psi_init), lindbladian)
                    pop0s.append(float(pops[0]))
                surv_prob.append(pop0s)
            lengths = np.append(lengths, new_lengths)
    epc = 0.5 * (1 - r)
    epg = 1 - ((1 - epc) ** (1 / 4))  # TODO: adjust to be mean length of
    return epg


@fid_reg_deco
def lindbladian_RB_left(
    U_dict: dict,
    gate: str,
    index,
    dims,
    proj: bool = False,
):
    return RB(U_dict, padding="left")


@fid_reg_deco
def lindbladian_RB_right(U_dict: dict, gate: str, index, dims, proj: bool):
    return RB(U_dict, padding="right")


@fid_reg_deco
def leakage_RB(
    U_dict,
    min_length: int = 5,
    max_length: int = 500,
    num_lengths: int = 20,
    num_seqs: int = 30,
    logspace=False,
    lindbladian=False,
):
    gate = list(U_dict.keys())[0]
    U = U_dict[gate]
    dim = int(U.shape[0])
    psi_init = tf.constant(basis(dim, 0), dtype=tf.complex128)
    if logspace:
        lengths = np.rint(
            np.logspace(np.log10(min_length), np.log10(max_length), num=num_lengths)
        ).astype(int)
    else:
        lengths = np.rint(np.linspace(min_length, max_length, num=num_lengths)).astype(
            int
        )
    comp_surv = []
    surv_prob = []
    for L in lengths:
        seqs = single_length_RB(num_seqs, L)
        Us = evaluate_sequences(U_dict, seqs)
        pop0s = []
        pop_comps = []
        for U in Us:
            pops = populations(tf.matmul(U, psi_init), lindbladian)
            pop0s.append(float(pops[0]))
            pop_comps.append(float(pops[0]) + float(pops[1]))
        surv_prob.append(pop0s)
        comp_surv.append(pop_comps)

    def RB_leakage(len, r_leak, A_leak, B_leak):
        return A_leak + B_leak * r_leak ** (len)

    bounds = (0, 1)
    init_guess = [0.9, 0.5, 0.5]
    fitted = False
    while not fitted:
        try:
            comp_means = np.mean(comp_surv, axis=1)
            comp_stds = np.std(comp_surv, axis=1) / np.sqrt(len(comp_surv[0]))
            solution, cov = curve_fit(
                RB_leakage,
                lengths,
                comp_means,
                sigma=comp_stds,
                bounds=bounds,
                p0=init_guess,
            )
            r_leak, A_leak, B_leak = solution
            fitted = True
        except Exception as message:
            print(message)
            if logspace:
                new_lengths = np.rint(
                    np.logspace(
                        np.log10(max_length + min_length),
                        np.log10(max_length * 2),
                        num=num_lengths,
                    )
                ).astype(int)
            else:
                new_lengths = np.rint(
                    np.linspace(
                        max_length + min_length, max_length * 2, num=num_lengths
                    )
                ).astype(int)
            max_length = max_length * 2
            for L in new_lengths:
                seqs = single_length_RB(num_seqs, L)
                Us = evaluate_sequences(U_dict, seqs)
                pop0s = []
                pop_comps = []
                for U in Us:
                    pops = populations(tf.matmul(U, psi_init), lindbladian)
                    pop0s.append(float(pops[0]))
                    pop_comps.append(float(pops[0]))
                surv_prob.append(pop0s)
                comp_surv.append(pop_comps)
            lengths = np.append(lengths, new_lengths)

    def RB_surv(len, r, A, C):
        return A + B_leak * r_leak ** (len) + C * r ** (len)

    bounds = (0, 1)
    init_guess = [0.9, 0.5, 0.5]

    fitted = False
    while not fitted:
        try:
            surv_means = np.mean(surv_prob, axis=1)
            surv_stds = np.std(surv_prob, axis=1) / np.sqrt(len(surv_prob[0]))
            solution, cov = curve_fit(
                RB_surv,
                lengths,
                surv_means,
                sigma=surv_stds,
                bounds=bounds,
                p0=init_guess,
            )
            r, A, C = solution
            fitted = True
        except Exception as message:
            print(message)
            if logspace:
                new_lengths = np.rint(
                    np.logspace(
                        np.log10(max_length + min_length),
                        np.log10(max_length * 2),
                        num=num_lengths,
                    )
                ).astype(int)
            else:
                new_lengths = np.rint(
                    np.linspace(
                        max_length + min_length, max_length * 2, num=num_lengths
                    )
                ).astype(int)
            max_length = max_length * 2
            for L in new_lengths:
                seqs = single_length_RB(num_seqs, L)
                Us = evaluate_sequences(U_dict, seqs)
                pop0s = []
                pop_comps = []
                for U in Us:
                    pops = populations(tf.matmul(U, psi_init), lindbladian)
                    pop0s.append(float(pops[0]))
                    pop_comps.append(float(pops[0]))
                surv_prob.append(pop0s)
                comp_surv.append(pop_comps)
            lengths = np.append(lengths, new_lengths)

    leakage = (1 - A_leak) * (1 - r_leak)
    seepage = A_leak * (1 - r_leak)
    fid = 0.5 * (r + 1 - leakage)
    epc = 1 - fid
    return epc, leakage, seepage, r_leak, A_leak, B_leak, r, A, C


@fid_reg_deco
def orbit_infid(
    U_dict,
    RB_number: int = 30,
    RB_length: int = 20,
    lindbladian=False,
    shots: int = None,
    seqs=None,
    noise=None,
):
    if not seqs:
        seqs = single_length_RB(RB_number=RB_number, RB_length=RB_length)
    Us = evaluate_sequences(U_dict, seqs)
    infids = []
    for U in Us:
        dim = int(U.shape[0])
        psi_init = tf.constant(basis(dim, 0), dtype=tf.complex128)
        psi_actual = tf.matmul(U, psi_init)
        pop0 = tf_abs(psi_actual[0]) ** 2
        p1 = 1 - pop0
        if shots:
            vals = tf.keras.backend.random_binomial(
                [shots],
                p=p1,
                dtype=tf.float64,
            )
            # if noise:
            #     vals = vals + (np.random.randn(shots) * noise)
            infid = tf.reduce_mean(vals)
        else:
            infid = p1
            # if noise:
            #     infid = infid + (np.random.randn() * noise)
        if noise:
            infid = infid + (np.random.randn() * noise)

        infids.append(infid)
    return tf_ave(infids)



"""
integration testing module for tunable coupler element
and line specific chain of signal generation.
"""

# System imports
import copy
import pickle
import pytest
import numpy as np

# Main C3 objects
from c3.c3objs import Quantity as Qty
from c3.parametermap import ParameterMap as PMap
from c3.experiment import Experiment as Exp
from c3.system.model import Model as Mdl
from c3.generator.generator import Generator as Gnr

# Building blocks
import c3.generator.devices as devices
import c3.signal.gates as gates
import c3.system.chip as chip
import c3.signal.pulse as pulse

# Libs and helpers
import c3.libraries.hamiltonians as hamiltonians
import c3.libraries.envelopes as envelopes


lindblad = False
dressed = True
q1_lvls = 3
q2_lvls = 3
tc_lvls = 3
freq_q1 = 6.189e9
freq_q2 = 5.089e9
freq_tc = 8.1e9
phi_0_tc = 10
fluxpoint = phi_0_tc * 0.23
d = 0.36

anhar_q1 = -286e6
anhar_q2 = -310e6
anhar_TC = -235e6
coupling_strength_q1tc = 142e6
coupling_strength_q2tc = 116e6
coupling_strength_q1q2 = 0 * 1e6

t1_q1 = 23e-6
t1_q2 = 70e-6
t1_tc = 15e-6
t2star_q1 = 27e-6
t2star_q2 = 50e-6
t2star_tc = 7e-6
init_temp = 0.06
v2hz = 1e9
t_final = 10e-9  # Time for single qubit gates
sim_res = 100e9
awg_res = 2.4e9

cphase_time = 100e-9  # Two qubit gate
flux_freq = 829 * 1e6
offset = 0 * 1e6
fluxamp = 0.1 * phi_0_tc
t_down = cphase_time - 5e-9
xy_angle = 0.3590456701578104
framechange_q1 = 0.725 * np.pi
framechange_q2 = 1.221 * np.pi

# ### MAKE MODEL
q1 = chip.Qubit(
    name="Q1",
    desc="Qubit 1",
    freq=Qty(value=freq_q1, min_val=5.0e9, max_val=8.0e9, unit="Hz 2pi"),
    anhar=Qty(value=anhar_q1, min_val=-380e6, max_val=-120e6, unit="Hz 2pi"),
    hilbert_dim=q1_lvls,
    t1=Qty(value=t1_q1, min_val=5e-6, max_val=90e-6, unit="s"),
    t2star=Qty(value=t2star_q1, min_val=10e-6, max_val=90e-6, unit="s"),
    temp=Qty(value=init_temp, min_val=0.0, max_val=0.12, unit="K"),
)
q2 = chip.Qubit(
    name="Q2",
    desc="Qubit 2",
    freq=Qty(value=freq_q2, min_val=5.0e9, max_val=8.0e9, unit="Hz 2pi"),
    anhar=Qty(value=anhar_q2, min_val=-380e6, max_val=-120e6, unit="Hz 2pi"),
    hilbert_dim=q2_lvls,
    t1=Qty(value=t1_q2, min_val=5e-6, max_val=90e-6, unit="s"),
    t2star=Qty(value=t2star_q2, min_val=10e-6, max_val=90e-6, unit="s"),
    temp=Qty(value=init_temp, min_val=0.0, max_val=0.12, unit="K"),
)
tc_at = chip.Transmon(
    name="TC",
    desc="Tunable Coupler",
    freq=Qty(value=freq_tc, min_val=0.0e9, max_val=10.0e9, unit="Hz 2pi"),
    phi=Qty(
        value=fluxpoint, min_val=-5.0 * phi_0_tc, max_val=5.0 * phi_0_tc, unit="Wb"
    ),
    phi_0=Qty(
        value=phi_0_tc, min_val=phi_0_tc * 0.9, max_val=phi_0_tc * 1.1, unit="Wb"
    ),
    d=Qty(value=d, min_val=d * 0.9, max_val=d * 1.1, unit=""),
    hilbert_dim=tc_lvls,
    anhar=Qty(value=anhar_TC, min_val=-380e6, max_val=-120e6, unit="Hz 2pi"),
    t1=Qty(value=t1_tc, min_val=1e-6, max_val=90e-6, unit="s"),
    t2star=Qty(value=t2star_tc, min_val=1e-6, max_val=90e-6, unit="s"),
    temp=Qty(value=init_temp, min_val=0.0, max_val=0.12, unit="K"),
)
q1tc = chip.Coupling(
    name="Q1-TC",
    desc="Coupling qubit 1 to tunable coupler",
    connected=["Q1", "TC"],
    strength=Qty(
        value=coupling_strength_q1tc, min_val=0 * 1e4, max_val=200e6, unit="Hz 2pi"
    ),
    hamiltonian_func=hamiltonians.int_XX,
)
q2tc = chip.Coupling(
    name="Q2-TC",
    desc="Coupling qubit 2 to t×unable coupler",
    connected=["Q2", "TC"],
    strength=Qty(
        value=coupling_strength_q2tc, min_val=0 * 1e4, max_val=200e6, unit="Hz 2pi"
    ),
    hamiltonian_func=hamiltonians.int_XX,
)
q1q2 = chip.Coupling(
    name="Q1-Q2",
    desc="Coupling qubit 1 to qubit 2",
    connected=["Q1", "Q2"],
    strength=Qty(
        value=coupling_strength_q1q2, min_val=0 * 1e4, max_val=200e6, unit="Hz 2pi"
    ),
    hamiltonian_func=hamiltonians.int_XX,
)
drive_q1 = chip.Drive(
    name="Q1",
    desc="Drive on Q1",
    connected=["Q1"],
    hamiltonian_func=hamiltonians.x_drive,
)
drive_q2 = chip.Drive(
    name="Q2",
    desc="Drive on Q2",
    connected=["Q2"],
    hamiltonian_func=hamiltonians.x_drive,
)
flux = chip.Drive(
    name="TC",
    desc="Flux drive/control on tunable couler",
    connected=["TC"],
    hamiltonian_func=hamiltonians.z_drive,
)
phys_components = [tc_at, q1, q2]
line_components = [flux, q1tc, q2tc, q1q2, drive_q1, drive_q2]

model = Mdl(phys_components, line_components, [])
model.set_lindbladian(lindblad)
model.set_dressed(dressed)

# ### MAKE GENERATOR
lo = devices.LO(name="lo", resolution=sim_res)
awg = devices.AWG(name="awg", resolution=awg_res)
dig_to_an = devices.DigitalToAnalog(name="dac", resolution=sim_res)
resp = devices.Response(
    name="resp",
    rise_time=Qty(value=0.3e-9, min_val=0.05e-9, max_val=0.6e-9, unit="s"),
    resolution=sim_res,
)
mixer = devices.Mixer(name="mixer")
fluxbias = devices.FluxTuning(
    name="fluxbias",
    phi_0=Qty(
        value=phi_0_tc, min_val=0.9 * phi_0_tc, max_val=1.1 * phi_0_tc, unit="Wb"
    ),
    phi=Qty(
        value=fluxpoint, min_val=-1.0 * phi_0_tc, max_val=1.0 * phi_0_tc, unit="Wb"
    ),
    omega_0=Qty(
        value=freq_tc, min_val=0.9 * freq_tc, max_val=1.1 * freq_tc, unit="Hz 2pi"
    ),
    d=Qty(value=d, min_val=d * 0.9, max_val=d * 1.1, unit=""),
    anhar=Qty(value=anhar_q1, min_val=-380e6, max_val=-120e6, unit="Hz 2pi"),
)
v_to_hz = devices.VoltsToHertz(
    name="v2hz",
    V_to_Hz=Qty(value=v2hz, min_val=0.9 * v2hz, max_val=1.1 * v2hz, unit="Hz 2pi/V"),
)
device_dict = {
    dev.name: dev for dev in [lo, awg, mixer, dig_to_an, resp, v_to_hz, fluxbias]
}
generator = Gnr(
    devices=device_dict,
    chains={
        "TC": ["lo", "awg", "dac", "resp", "mixer", "fluxbias"],
        "Q1": ["lo", "awg", "dac", "resp", "mixer", "v2hz"],
        "Q2": ["lo", "awg", "dac", "resp", "mixer", "v2hz"],
    },
)

# ### MAKE GATESET
nodrive_env = pulse.Envelope(name="no_drive", params={}, shape=envelopes.no_drive)
carrier_parameters = {
    "freq": Qty(value=freq_q1, min_val=0e9, max_val=10e9, unit="Hz 2pi"),
    "framechange": Qty(value=0.0, min_val=-3 * np.pi, max_val=5 * np.pi, unit="rad"),
}
carr_q1 = pulse.Carrier(
    name="carrier", desc="Frequency of the local oscillator", params=carrier_parameters
)
carr_q2 = copy.deepcopy(carr_q1)
carr_q2.params["freq"].set_value(freq_q2)
carr_tc = copy.deepcopy(carr_q1)
carr_tc.params["freq"].set_value(flux_freq)


flux_params = {
    "amp": Qty(value=fluxamp, min_val=0.0, max_val=5, unit="V"),
    "t_final": Qty(
        value=cphase_time,
        min_val=0.5 * cphase_time,
        max_val=1.5 * cphase_time,
        unit="s",
    ),
    "t_up": Qty(
        value=5 * 1e-9, min_val=0.0 * cphase_time, max_val=0.5 * cphase_time, unit="s"
    ),
    "t_down": Qty(
        value=t_down, min_val=0.5 * cphase_time, max_val=1.0 * cphase_time, unit="s"
    ),
    "risefall": Qty(
        value=5 * 1e-9, min_val=0.0 * cphase_time, max_val=1.0 * cphase_time, unit="s"
    ),
    "freq_offset": Qty(
        value=offset, min_val=-50 * 1e6, max_val=50 * 1e6, unit="Hz 2pi"
    ),
    "xy_angle": Qty(
        value=xy_angle, min_val=-0.5 * np.pi, max_val=2.5 * np.pi, unit="rad"
    ),
}
flux_env = pulse.Envelope(
    name="flux",
    desc="Flux bias for tunable coupler",
    params=flux_params,
    shape=envelopes.flattop,
)
CRZp = gates.Instruction(
    name="Id:CRZp", t_start=0.0, t_end=cphase_time, channels=["Q1", "Q2", "TC"]
)
CRZp.add_component(flux_env, "TC")
CRZp.add_component(carr_tc, "TC")
CRZp.add_component(nodrive_env, "Q1")
CRZp.add_component(carr_q1, "Q1")
CRZp.comps["Q1"]["carrier"].params["framechange"].set_value(framechange_q1)
CRZp.add_component(nodrive_env, "Q2")
CRZp.add_component(carr_q2, "Q2")
CRZp.comps["Q2"]["carrier"].params["framechange"].set_value(framechange_q2)


# ### MAKE EXPERIMENT
parameter_map = PMap(instructions=[CRZp], model=model, generator=generator)
exp = Exp(pmap=parameter_map)

##### TESTING ######

with open("test/tunable_coupler_data.pickle", "rb") as filename:
    data = pickle.load(filename)


@pytest.mark.integration
def test_coupler_frequency() -> None:
    coupler_01 = np.abs(
        np.abs(model.eigenframe[model.state_labels.index((0, 0, 0))])
        - np.abs(model.eigenframe[model.state_labels.index((1, 0, 0))])
    )
    rel_diff = np.abs((coupler_01 - data["coupler_01"]) / data["coupler_01"])
    assert rel_diff < 1e-12


@pytest.mark.integration
def test_coupler_anahrmonicity() -> None:
    coupler_12 = np.abs(
        np.abs(model.eigenframe[model.state_labels.index((1, 0, 0))])
        - np.abs(model.eigenframe[model.state_labels.index((2, 0, 0))])
    )
    rel_diff = np.abs((coupler_12 - data["coupler_12"]) / data["coupler_12"])
    assert rel_diff < 1e-12


@pytest.mark.integration
def test_energy_levels() -> None:
    model = parameter_map.model
    parameter_map.set_parameters([0.0], [[["TC-phi"]]])
    model.update_model()
    labels = [
        model.state_labels[indx]
        for indx in np.argsort(np.abs(model.eigenframe) / 2 / np.pi / 1e9)
    ]
    product_basis = []
    dressed_basis = []
    ordered_basis = []
    transforms = []
    steps = 101
    min_ratio = -0.10
    max_ratio = 0.7
    flux_ratios = np.linspace(min_ratio, max_ratio, steps, endpoint=True)
    for flux_ratio in flux_ratios:
        flux_bias = flux_ratio * phi_0_tc
        parameter_map.set_parameters(
            [flux_bias, 0.0, 0.0, 0.0],
            [
                [["TC-phi"]],
                [["Q1-TC-strength"]],
                [["Q2-TC-strength"]],
                [["Q1-Q2-strength"]],
            ],
        )
        model.update_model()
        product_basis.append(
            [
                model.eigenframe[model.state_labels.index(label)] / 2 / np.pi / 1e9
                for label in labels
            ]
        )
        parameter_map.set_parameters(
            [coupling_strength_q1tc, coupling_strength_q2tc, coupling_strength_q1q2],
            [[["Q1-TC-strength"]], [["Q2-TC-strength"]], [["Q1-Q2-strength"]]],
        )
        model.update_model()
        ordered_basis.append(
            [
                model.eigenframe[model.state_labels.index(label)] / 2 / np.pi / 1e9
                for label in labels
            ]
        )
        parameter_map.model.update_dressed(ordered=False)
        dressed_basis.append(
            [
                model.eigenframe[model.state_labels.index(label)] / 2 / np.pi / 1e9
                for label in model.state_labels
            ]
        )
        transforms.append(
            np.array(
                [
                    np.real(model.transform[model.state_labels.index(label)])
                    for label in labels
                ]
            )
        )
    parameter_map.set_parameters([fluxpoint], [[["TC-phi"]]])
    model.update_model()
    dressed_basis = np.array(dressed_basis)
    ordered_basis = np.array(ordered_basis)
    product_basis = np.array(product_basis)
    print((np.abs(product_basis - data["product_basis"]) < 1).all())
    assert (np.abs(product_basis - data["product_basis"]) < 1).all()
    assert (np.abs(ordered_basis - data["ordered_basis"]) < 1).all()
    # Dressed basis might change at avoided crossings depending on how we
    # decide to deal with it. Atm no state with largest probability is chosen.
    assert (np.abs(dressed_basis - data["dressed_basis"]) < 1).all()


@pytest.mark.slow
@pytest.mark.integration
def test_dynamics_CPHASE() -> None:
    # Dynamics (closed system)
    exp.set_opt_gates(["Id:CRZp"])
    exp.get_gates()
    dUs = []
    for indx in range(len(exp.dUs["Id:CRZp"])):
        if indx % 50 == 0:
            dUs.append(exp.dUs["Id:CRZp"][indx].numpy())
    dUs = np.array(dUs)
    assert (np.abs(np.real(dUs) - np.real(data["dUs"])) < 1e-8).all()
    assert (np.abs(np.imag(dUs) - np.imag(data["dUs"])) < 1e-8).all()
    assert (np.abs(np.abs(dUs) - np.abs(data["dUs"])) < 1e-8).all()


@pytest.mark.integration
@pytest.mark.slow
@pytest.mark.heavy
@pytest.mark.skip(reason="takes way too long")
def test_dynamics_CPHASE_lindblad() -> None:
    # Dynamics (open system)
    exp.pmap.model.set_lindbladian(True)
    U_dict = exp.get_gates()
    # saved U_super currenlty likely to be wrong. Not recomputed in last dataset.
    U_super = U_dict["Id:CRZp"]
    assert (np.abs(np.real(U_super) - np.real(data["U_super"])) < 1e-8).all()
    assert (np.abs(np.imag(U_super) - np.imag(data["U_super"])) < 1e-8).all()
    assert (np.abs(np.abs(U_super) - np.abs(data["U_super"])) < 1e-8).all()
    assert (np.abs(np.angle(U_super) - np.angle(data["U_super"])) < 1e-8).all()


@pytest.mark.integration
def test_separate_chains() -> None:
    assert generator.chains["Q2"] == generator.chains["Q1"]
    assert generator.chains["TC"] != generator.chains["Q1"]
    assert generator.chains["TC"] != generator.chains["Q2"]


@pytest.mark.slow
@pytest.mark.integration
def test_flux_signal() -> None:
    instr = exp.pmap.instructions["Id:CRZp"]
    signal = exp.pmap.generator.generate_signals(instr)
    awg = exp.pmap.generator.devices["awg"]
    # mixer = exp.pmap.generator.devices["mixer"]
    channel = "TC"
    tc_signal = signal[channel]["values"].numpy()
    tc_ts = signal[channel]["ts"].numpy()
    tc_awg_I = awg.signal[channel]["inphase"].numpy()
    tc_awg_Q = awg.signal[channel]["quadrature"].numpy()
    tc_awg_ts = awg.signal[channel]["ts"].numpy()
    rel_diff = np.abs((tc_signal - data["tc_signal"]) / np.max(data["tc_signal"]))
    assert (rel_diff < 1e-12).all()
    rel_diff = np.abs((tc_ts - data["tc_ts"]) / np.max(data["tc_ts"]))
    assert (rel_diff < 1e-12).all()
    rel_diff = np.abs((tc_awg_I - data["tc_awg_I"]) / np.max(data["tc_awg_I"]))
    assert (rel_diff < 1e-12).all()
    rel_diff = np.abs((tc_awg_Q - data["tc_awg_Q"]) / np.max(data["tc_awg_Q"]))
    assert (rel_diff < 1e-12).all()
    rel_diff = np.abs((tc_awg_ts - data["tc_awg_ts"]) / np.max(data["tc_awg_ts"]))
    assert (rel_diff < 1e-12).all()


@pytest.mark.unit
def test_FluxTuning():
    flux_tune = devices.FluxTuning(
        name="flux_tune",
        phi_0=Qty(phi_0_tc),
        phi=Qty(value=0, min_val=-phi_0_tc, max_val=phi_0_tc),
        omega_0=Qty(freq_tc),
        anhar=Qty(anhar_TC),
        d=Qty(d),
    )

    transmon = chip.Transmon(
        name="transmon",
        hilbert_dim=3,
        freq=Qty(freq_tc),
        phi=Qty(value=0, min_val=-1.5 * phi_0_tc, max_val=1.5 * phi_0_tc),
        phi_0=Qty(phi_0_tc),
        d=Qty(d),
        anhar=Qty(anhar_TC),
    )

    bias_phis = [0, 0.2]
    phis = np.linspace(-1, 1, 10) * phi_0_tc

    for bias_phi in bias_phis:
        flux_tune.params["phi"].set_value(bias_phi)
        signal = {"ts": np.linspace(0, 1, 10), "values": phis}
        signal_out = flux_tune.process(None, None, signal)
        flux_tune_frequencies = signal_out["values"].numpy()

        transmon_frequencies = []
        transmon.params["phi"].set_value(bias_phi)
        bias_freq = transmon.get_freq()
        for phi in phis + bias_phi:
            transmon.params["phi"].set_value(phi)
            transmon_frequencies.append(transmon.get_freq())
        transmon_diff_freq = np.array(transmon_frequencies) - bias_freq

        assert (
            np.max(np.abs(flux_tune_frequencies - transmon_diff_freq)) < 1e-15 * freq_tc
        )
